import BrtcOHOSSdk from 'libbrtcohossdk.so';
import hilog from '@ohos.hilog';
import { BusinessError } from '@kit.BasicServicesKit';
import * as brtc from "./BrtcDefines";
import { BrtcListener } from "./BrtcEventHandler"
import BrtcDeviceInfo from "./BrtcDeviceInfo"
import BrtcDeviceManagerImpl from './BrtcDeviceManagerImpl'
import { BrtcNetworkQosParam, BrtcVideoRotation, BrtcVideoStreamType } from './BrtcDefines';
import { AbilityLifecycleCallback, UIAbility } from '@kit.AbilityKit';
import { window } from '@kit.ArkUI';
import { common } from '@kit.AbilityKit';
import { fileIo } from '@kit.CoreFileKit';
import { CardType } from '@kit.VisionKit';
import { RootConfig } from '../ExperimentalApiDataModel';

let engineImpl: BrtcEngineImpl | null
type Callback = (...args: any[]) => void

export class BrtcEngineImpl {
  private TAG: string = "BrtcEngineImpl";
  private nativeBrtcEngine: any;
  private events: Record<string, Callback[]>;
  private deviceManager: BrtcDeviceManagerImpl;
  private applicationContext: Object | null = null;

  private lifecycleCallback: AbilityLifecycleCallback = {
    onAbilityCreate(ability) {
      console.info('AbilityLifecycleCallback onAbilityCreate');
    },
    onWindowStageCreate: (ability: UIAbility, windowStage: window.WindowStage): void => {
      console.info('AbilityLifecycleCallback onWindowStageCreate');
    },
    onWindowStageActive: (ability: UIAbility, windowStage: window.WindowStage): void => {
      console.info('AbilityLifecycleCallback onWindowStageActive');
    },
    onWindowStageInactive: (ability: UIAbility, windowStage: window.WindowStage): void => {
      console.info('AbilityLifecycleCallback onWindowStageInactive');
    },
    onWindowStageDestroy: (ability: UIAbility, windowStage: window.WindowStage): void => {
      console.info('AbilityLifecycleCallback onWindowStageDestroy');
    },
    onAbilityDestroy: (ability: UIAbility): void => {
      console.info('AbilityLifecycleCallback onAbilityDestroy');
    },
    onAbilityForeground: (ability: UIAbility): void => {
      console.info('AbilityLifecycleCallback onAbilityForeground');
      BrtcOHOSSdk.backgroundStateChanged(false);
    },
    onAbilityBackground: (ability: UIAbility): void => {
      console.info('AbilityLifecycleCallback onAbilityBackground');
      BrtcOHOSSdk.backgroundStateChanged(true);
    },
    onAbilityContinue: (ability: UIAbility): void => {
      console.info('AbilityLifecycleCallback onAbilityContinue');
    }
  };

  async extractRawfile() {
    if (this.applicationContext == null) {
      return;
    }
    let ctx = this.applicationContext as common.ApplicationContext;
    let resourceMgr = ctx.resourceManager;
    try {
      resourceMgr.getRawFileContent("gvalue2.pcm", (error: BusinessError, value: Uint8Array) => {
        if (error != null) {
          console.error("Failed to extract gvalue2.pcm, error is " + error);
        } else {
          // for example: /data/storage/el2/base/files/gvalue2.pcm
          // realpath:    /data/app/el2/100/base/com.baijiayun.brtcohos/files/gvalue2.pcm
          let path: string = ctx.filesDir + "/gvalue2.pcm";
          let file = fileIo.openSync(path, fileIo.OpenMode.READ_WRITE | fileIo.OpenMode.CREATE);
          fileIo.write(file.fd, value.buffer).then((writeLen) => {
            fileIo.closeSync(file);
            const config: RootConfig = {
              'brtc.audio.config': {
                audio_omlsa_file_path: path
              }
            };
            this.callExperimentalApi(JSON.stringify(config));
          }).catch((err: BusinessError) => {
            console.error("Failed to write gvalue2.pcm to file, error is " + err);
          })
        }
      });
    } catch (error) {
      let code = (error as BusinessError).code;
      let message = (error as BusinessError).message;
      console.error(`callback getRawFileContent failed, error code: ${code}, message: ${message}.`);
    }
  }

  constructor() {
    this.nativeBrtcEngine = null
    this.events = {}
    this.deviceManager = new BrtcDeviceManagerImpl()
  }

  static getInstance(): BrtcEngineImpl {
    if (engineImpl) {
      return engineImpl;
    } else {
      throw new Error("Failed to get BRTC engine instance, please create first")
    }
  }

  static async createEngine(config: brtc.BrtcEngineAdvancedConfig | null): Promise<BrtcEngineImpl | null> {
    if (engineImpl) {
      return engineImpl as BrtcEngineImpl;
    } else {
      engineImpl = new BrtcEngineImpl;
      const result = engineImpl.initContext(config);
      if (result) {
        return engineImpl;
      } else {
        hilog.info(0x0000, "BRTC", "Failed to init BRTCEngine context");
        engineImpl = null;
        return null;
      }
    }
  }

  static destroyEngine() {
    if (engineImpl) {
      return engineImpl.unInitContext();
    }
  }

  static getSdkVersion(): string {
    return BrtcOHOSSdk.getSdkVersion();
  }

  static setConsoleEnabled(enabled: boolean): void {
    return BrtcOHOSSdk.setConsoleEnabled(enabled);
  }

  static setLogLevel(level: brtc.BrtcLogLevel): void {
    return BrtcOHOSSdk.setLogLevel(level);
  }

  static setLogPath(path: string): void {
    return BrtcOHOSSdk.setLogPath(path);
  }

  callExperimentalApi(jsonStr: string): void {
    this.nativeBrtcEngine?.callExperimentalApi(jsonStr);
  }

  setDefaultStreamRecvMode(autoRecvAudio: boolean, autoRecvVideo: boolean): void {
    this.nativeBrtcEngine?.setDefaultStreamRecvMode(autoRecvAudio, autoRecvVideo);
  }

  setNetworkQosParam(qosParam: BrtcNetworkQosParam): void {
    if (qosParam == null || qosParam == undefined) {
      return;
    }
    this.nativeBrtcEngine?.setNetworkQosParam(qosParam);
  }

  enableSmallVideoStream(enable: boolean, encParam: brtc.BrtcVideoEncParam): void {
    if (encParam == null || encParam == undefined) {
      return;
    }
    this.nativeBrtcEngine?.enableSmallVideoStream(enable, encParam);
  }

  setVideoEncoderParam(encParam: brtc.BrtcVideoEncParam): void {
    if (encParam == null || encParam == undefined) {
      return;
    }
    this.nativeBrtcEngine?.setVideoEncoderParam(encParam);
  }

  setVideoEncoderMirror(mirror: boolean): void {
    this.nativeBrtcEngine?.setVideoEncoderMirror(mirror);
  }

  setVideoEncoderRotation(rotation: BrtcVideoRotation): void {
    this.nativeBrtcEngine?.setVideoEncoderRotation(rotation);
  }

  enterRoom(roomParams: brtc.BrtcRoomParams): void {
    this.nativeBrtcEngine?.enterRoom(roomParams)
  }

  exitRoom(): void {
    this.nativeBrtcEngine?.exitRoom()
  }

  startLocalPreview(useFront: boolean): void {
    this.nativeBrtcEngine?.startLocalPreview(useFront)
  }

  stopLocalPreview(): void {
    this.nativeBrtcEngine?.stopLocalPreview()
  }

  startLocalAudio(quality: brtc.BrtcAudioQuality): void {
    this.nativeBrtcEngine?.startLocalAudio(quality)
  }

  stopLocalAudio(): void {
    this.nativeBrtcEngine?.stopLocalAudio()
  }

  muteLocalAudio(mute: boolean): void {
    this.nativeBrtcEngine?.muteLocalAudio(mute);
  }

  muteLocalVideo(streamType: brtc.BrtcVideoStreamType, mute: boolean): void {
    this.nativeBrtcEngine?.muteLocalVideo(streamType, mute);
  }

  muteRemoteAudio(userId: string, mute: boolean): void {
    this.nativeBrtcEngine?.muteRemoteAudio(userId, mute);
  }

  muteRemoteVideo(userId: string, streamType: brtc.BrtcVideoStreamType, mute: boolean): void {
    this.nativeBrtcEngine?.muteRemoteVideo(userId, streamType, mute);
  }

  muteAllRemoteAudio(mute: boolean): void {
    this.nativeBrtcEngine?.muteAllRemoteAudio(mute);
  }

  startScreenCapture(xComponentSurfaceId:string, streamType: brtc.BrtcVideoStreamType,
    encParam: brtc.BrtcVideoEncParam): void {
    this.nativeBrtcEngine?.startScreenCapture(xComponentSurfaceId, streamType, encParam);
  }

  stopScreenCapture(): void {
    this.nativeBrtcEngine?.stopScreenCapture();
  }

  getDeviceManager(): BrtcDeviceManagerImpl {
    return this.deviceManager;
  }

  startRemoteView(userId: string, streamType: BrtcVideoStreamType) {
    this.nativeBrtcEngine?.startRemoteView(userId, streamType);
  }

  stopRemoteView(userId: string, streamType: BrtcVideoStreamType) {
    this.nativeBrtcEngine?.stopRemoteView(userId, streamType);
  }

  setRemoteVideoStreamType(userId: string, streamType: BrtcVideoStreamType) {
    this.nativeBrtcEngine?.setRemoteVideoStreamType(userId, streamType);
  }

  setRenderParams(userId: string, streamType: brtc.BrtcVideoStreamType, params: brtc.BrtcRenderParams) {
    this.nativeBrtcEngine?.setRenderParams(userId, streamType, params);
  }

  on<EventType extends keyof BrtcListener>(event: EventType, callback: BrtcListener[EventType]) {
    if (!this.events[event]) {
      // 首次初始化一个空数组，用于容纳多个 callback
      this.events[event] = [];
    }
    this.events[event].push(callback);
  }

  off<EventType extends keyof BrtcListener>(event: EventType, callback: BrtcListener[EventType]) {
    const callbacks = this.events[event];
    if (callbacks) {
      this.events[event] = callbacks.filter(cb => cb != callback);
      if (this.events[event].length == 0) {
        delete this.events[event];
      }
    }
  }

  emit(eventName: string, ...args: any[]) {
    const callbacks = this.events[eventName];
    if (callbacks) {
      callbacks.forEach(callback => {
        callback(...args);
      })
    }
  }

  callEmit(...args: any[]) {
    try {
      let params = args.slice(1)
      if (args[0] == 'onStatistics') {
        // TODO: 将 args[1] (json 格式的 BRTCStatistics 字符串）转为 js 类对象，方便用户使用
      }
      this.emit(args[0], ...params)
    } catch (e) {
      let error: BusinessError = e as BusinessError;
      hilog.error(0x0000, this.TAG, 'call function: %{public}s exception, message: %{public}s', args[0], error.message);
    }
  }

  clearAllListener() {
    if (this.events != null) {
      this.events = {};
    }
  }

  private putExtraEngineConfig(configMap: Map<string, string>) {
    configMap.set('deviceInfo', BrtcDeviceInfo.getDeviceInfo());
  }

  private engineConfigToJson(engineConfig: brtc.BrtcEngineAdvancedConfig | null): string {
    // 如果外部没有设置，则初始化 engine config
    if (!engineConfig) {
      engineConfig = new brtc.BrtcEngineAdvancedConfig();
    }
    if (typeof engineConfig.config === 'undefined' || !engineConfig.config) {
      engineConfig.config = new Map();
    }

    // 添加内部必要的一些配置信息
    this.putExtraEngineConfig(engineConfig.config);

    let mapObj = {};
    for (let [key, value] of engineConfig.config) {
      mapObj[key] = value;
    }
    return JSON.stringify(mapObj);
  }

  private initContext(engineConfig: brtc.BrtcEngineAdvancedConfig | null): boolean {
    // 加载和初始化 node-api 模块
    const jsCallback = this.callEmit.bind(this);
    this.nativeBrtcEngine = new BrtcOHOSSdk.BrtcOhosEngine();
    this.applicationContext = engineConfig.applicationContext;

    if (this.nativeBrtcEngine) {
      let configJson = this.engineConfigToJson(engineConfig);
      hilog.info(0x0000, this.TAG, configJson);
      let result = this.nativeBrtcEngine.createEngine(configJson, jsCallback);
      hilog.error(0x0000, this.TAG, 'Create BRTC OHOS SDK engine return: %{public}s', result ? "true" : "false");
      this.deviceManager.setBrtcNativeEngine(this.nativeBrtcEngine);

      if (this.applicationContext != null) {
        (this.applicationContext as common.ApplicationContext).on('abilityLifecycle', this.lifecycleCallback);
        this.extractRawfile();
      }

      return result as boolean;
    }

    return false;
  }

  private unInitContext(): Promise<void> {
    return new Promise((resolve, reject) => {
      this.clearAllListener();
      this.deviceManager.setBrtcNativeEngine(null);
      if (this.nativeBrtcEngine) {
        hilog.error(0x0000, this.TAG, 'Destroy BRTC OHOS SDK engine');
        this.nativeBrtcEngine.destroyEngine();
      }
      this.nativeBrtcEngine = null;
      engineImpl = null;
      resolve();
    });
  }



} // class BrtcEngineImpl

export default BrtcEngineImpl;